package C::sparse::expand::EXPANSION_CMDLINE;
our @ISA = qw (C::sparse::expand);
package C::sparse::expand::EXPANSION_STREAM;
our @ISA = qw (C::sparse::expand);
package C::sparse::expand::EXPANSION_MACRO;
our @ISA = qw (C::sparse::expand);
package C::sparse::expand::EXPANSION_MACROARG;
our @ISA = qw (C::sparse::expand);
package C::sparse::expand::EXPANSION_CONCAT;
our @ISA = qw (C::sparse::expand);
package C::sparse::expand::EXPANSION_PREPRO;
our @ISA = qw (C::sparse::expand);

1;
